"""JSON manipulator."""
from __future__ import annotations

__all__: list[str] = ["Manipulator"]

import re
from copy import deepcopy
from decimal import Decimal, InvalidOperation
from operator import eq, ge, gt, le, lt, ne
from re import DOTALL, MULTILINE, VERBOSE, Match, RegexFlag
from typing import TYPE_CHECKING, Any

from jsonyx import JSONSyntaxError
from jsonyx.allow import NOTHING

if TYPE_CHECKING:
    from collections.abc import Callable, Container

    _Target = dict[Any, Any] | list[Any]
    _Node = tuple[_Target, Any]
    _MatchFunc = Callable[[str, int], Match[str] | None]
    _Operation = dict[str, Any]
    _Operator = Callable[[Any, Any], Any]


_FLAGS: RegexFlag = VERBOSE | MULTILINE | DOTALL

_match_idx: _MatchFunc = re.compile(r"-?0|-?[1-9][0-9]*", _FLAGS).match
_match_number: _MatchFunc = re.compile(
    r"""
    (-?0|-?[1-9][0-9]*) # integer
    (\.[0-9]+)?         # [frac]
    ([eE][-+]?[0-9]+)?  # [exp]
    """, _FLAGS,
).match
_match_slice: _MatchFunc = re.compile(
    r"""
    (-?0|-?[1-9][0-9]*)?       # [start]
    :                          # ":"
    (-?0|-?[1-9][0-9]*)?       # [stop]
    (?::(-?0|-?[1-9][0-9]*)?)? # [":" [step]]
    """, _FLAGS,
).match
_match_str_chunk: _MatchFunc = re.compile(r"[^'~]*", _FLAGS).match
_match_unquoted_key: _MatchFunc = re.compile(
    r"(?:\w+|[^\x00-\x7f]+)+", _FLAGS,
).match
_match_whitespace: _MatchFunc = re.compile(r"\ +", _FLAGS).match


def _errmsg(msg: str, query: str, start: int, end: int = 0) -> JSONSyntaxError:
    return JSONSyntaxError(msg, "<query>", query, start, end)


def _check_query_key(
    target: _Target, key: Any, *, allow_slice: bool = False,
) -> None:
    if isinstance(target, dict):
        if not isinstance(key, str):
            msg: str = f"Dict key must be str, not {type(key).__name__}"
            raise TypeError(msg)
    elif allow_slice:
        if isinstance(key, str):
            msg = f"List index must be int or slice, not {type(key).__name__}"
            raise TypeError(msg)
    elif not isinstance(key, int):
        msg = f"List index must be int, not {type(key).__name__}"
        raise TypeError(msg)


def _get_query_targets(
    node: _Node, *, relative: bool = False,
) -> list[_Target]:
    target, key = node
    _check_query_key(target, key, allow_slice=not relative)
    if isinstance(key, slice):
        new_targets: list[Any] = target[key]
    else:
        new_targets = [target[key]]

    for new_target in new_targets:
        if not isinstance(new_target, (dict, list)):
            msg: str = (
                f"Target must be dict or list, not {type(new_target).__name__}"
            )
            raise TypeError(msg)

    return new_targets


def _has_key(target: _Target, key: Any) -> bool:
    if isinstance(target, dict):
        return key in target

    return -len(target) <= key < len(target)  # type: ignore


def _scan_query_operator(query: str, end: int) -> tuple[_Operator | None, int]:
    if query[end:end + 2] == "<=":
        operator, end = le, end + 2
    elif query[end:end + 1] == "<":
        operator, end = lt, end + 1
    elif query[end:end + 2] == "==":
        operator, end = eq, end + 2
    elif query[end:end + 2] == "!=":
        operator, end = ne, end + 2
    elif query[end:end + 2] == ">=":
        operator, end = ge, end + 2
    elif query[end:end + 1] == ">":
        operator, end = gt, end + 1
    else:
        operator = None

    return operator, end


def _scan_query_string(s: str, end: int) -> tuple[str, int]:
    chunks: list[str] = []
    append_chunk: Callable[[str], None] = chunks.append
    str_idx: int = end - 1
    while True:
        if match := _match_str_chunk(s, end):
            end = match.end()
            append_chunk(match.group())

        try:
            terminator: str = s[end]
        except IndexError:
            msg: str = "Unterminated string"
            raise _errmsg(msg, s, str_idx, end) from None

        if terminator == "'":
            return "".join(chunks), end + 1

        end += 1
        try:
            esc: str = s[end]
        except IndexError:
            msg = "Expecting escaped character"
            raise _errmsg(msg, s, end) from None

        if esc not in {"'", "~"}:
            msg = "Invalid tilde escape"
            raise _errmsg(msg, s, end - 1, end + 1)

        end += 1
        append_chunk(esc)


def _asc_range(r: range) -> range:
    return r if r.step > 0 else r[::-1]


class Manipulator:
    """A configurable JSON manipulator.

    .. versionadded:: 2.0

    :param allow: the JSON deviations from :mod:`jsonyx.allow`
    :param use_decimal: use :class:`decimal.Decimal` instead of :class:`float`
    """

    def __init__(
        self, *, allow: Container[str] = NOTHING, use_decimal: bool = False,
    ) -> None:
        """Create a new JSON manipulator."""
        self._allow_nan_and_infinity: bool = "nan_and_infinity" in allow
        self._parse_float: Callable[
            [str], Decimal | float,
        ] = Decimal if use_decimal else float
        self._use_decimal: bool = use_decimal

    def _scan_query_value(self, s: str, idx: int = 0) -> tuple[Any, int]:
        try:
            nextchar: str = s[idx]
        except IndexError:
            msg: str = "Expecting value"
            raise _errmsg(msg, s, idx) from None

        value: Any
        if nextchar == "'":
            value, end = _scan_query_string(s, idx + 1)
        elif nextchar == "n" and s[idx:idx + 4] == "null":
            value, end = None, idx + 4
        elif nextchar == "t" and s[idx:idx + 4] == "true":
            value, end = True, idx + 4
        elif nextchar == "f" and s[idx:idx + 5] == "false":
            value, end = False, idx + 5
        elif number := _match_number(s, idx):
            integer, frac, exp = number.groups()
            end = number.end()
            if not frac and not exp:
                try:
                    value = int(integer)
                except ValueError:
                    msg = "Invalid number"
                    raise _errmsg(msg, s, idx, end) from None
            else:
                try:
                    value = self._parse_float(
                        integer + (frac or "") + (exp or ""),
                    )
                except InvalidOperation:
                    msg = "Invalid number"
                    raise _errmsg(msg, s, idx, end) from None
        elif nextchar == "I" and s[idx:idx + 8] == "Infinity":
            if not self._allow_nan_and_infinity:
                msg = "Infinity is not allowed"
                raise _errmsg(msg, s, idx, idx + 8)

            value, end = self._parse_float("Infinity"), idx + 8
        elif nextchar == "-" and s[idx:idx + 9] == "-Infinity":
            if not self._allow_nan_and_infinity:
                msg = "-Infinity is not allowed"
                raise _errmsg(msg, s, idx, idx + 9)

            value, end = self._parse_float("-Infinity"), idx + 9
        else:
            msg = "Expecting value"
            raise _errmsg(msg, s, idx)

        return value, end

    def _apply_filter(
        self, nodes: list[_Node], query: str, end: int,
    ) -> tuple[list[_Node], int]:
        while True:
            negate_filter: bool = query[end:end + 1] == "!"
            if negate_filter:
                end += 1

            filter_nodes, end = self._select_nodes(
                nodes, query, end, relative=True,
            )
            filtered_pairs: list[tuple[_Node, _Node]] = [
                (node, (filter_target, filter_key))
                for node, (filter_target, filter_key) in zip(
                    nodes, filter_nodes,
                )
                if _has_key(filter_target, filter_key) != negate_filter
            ]
            old_end: int = end
            if match := _match_whitespace(query, end):
                end = match.end()

            operator_idx: int = end
            operator, end = _scan_query_operator(query, end)
            if operator is None:
                nodes = [node for node, _filter_node in filtered_pairs]
            elif negate_filter:
                msg: str = "Unexpected operator"
                raise _errmsg(msg, query, operator_idx, end)
            else:
                if match := _match_whitespace(query, end):
                    end = match.end()

                value, end = self._scan_query_value(query, end)
                nodes = [
                    node
                    for node, (filter_target, filter_key) in filtered_pairs
                    if operator(filter_target[filter_key], value)
                ]
                old_end = end
                if match := _match_whitespace(query, end):
                    end = match.end()

            if query[end:end + 2] != "&&":
                return nodes, old_end

            end += 2
            if match := _match_whitespace(query, end):
                end = match.end()

    def _select_nodes(
        self,
        nodes: list[_Node],
        query: str,
        end: int = 0,
        *,
        allow_slice: bool = False,
        relative: bool = False,
    ) -> tuple[list[_Node], int]:
        if relative:
            if query[end:end + 1] != "@":
                msg: str = "Expecting a relative query"
                raise _errmsg(msg, query, end)
        elif query[end:end + 1] != "$":
            msg = "Expecting an absolute query"
            raise _errmsg(msg, query, end)

        end += 1
        while True:
            key: Any
            if query[end:end + 1] == "?":
                if relative:
                    msg = "Optional markers are not allowed in relative query"
                    raise _errmsg(msg, query, end, end + 1)

                end += 1
                for target, key in nodes:
                    _check_query_key(target, key, allow_slice=True)

                nodes = [
                    (target, key)
                    for (target, key) in nodes
                    if isinstance(key, slice) or _has_key(target, key)
                ]

            if (terminator := query[end:end + 1]) == ".":
                end += 1
                if (
                    match := _match_unquoted_key(query, end)
                ) and match.group().isidentifier():
                    key, end = match.group(), match.end()
                else:
                    msg = "Expecting property"
                    raise _errmsg(msg, query, end)

                nodes = [
                    (target, key)
                    for node in nodes
                    for target in _get_query_targets(node, relative=relative)
                ]
            elif terminator == "{":
                end += 1
                nodes = [
                    (target, new_key)
                    for target, key in nodes
                    for new_key in (
                        _asc_range(range(len(target))[key])
                        if isinstance(key, slice) else
                        [key]
                    )
                ]
                if relative:
                    msg = "Conditions are not allowed in relative query"
                    raise _errmsg(msg, query, end)

                nodes, end = self._apply_filter(nodes, query, end)
                if query[end:end + 1] != "}":
                    msg = "Expecting a closing bracket"
                    raise _errmsg(msg, query, end)

                end += 1
            elif terminator == "[":
                end += 1
                targets: list[_Target] = [
                    target
                    for node in nodes
                    for target in _get_query_targets(node, relative=relative)
                ]
                if match := _match_slice(query, end):
                    (start, stop, step), end = match.groups(), match.end()
                    try:
                        if start is not None:
                            start = int(start)
                    except ValueError:
                        msg = "Start is too big"
                        raise _errmsg(
                            msg, query, match.start(1), match.end(1),
                        ) from None

                    try:
                        if stop is not None:
                            stop = int(stop)
                    except ValueError:
                        msg = "Stop is too big"
                        raise _errmsg(
                            msg, query, match.start(2), match.end(2),
                        ) from None

                    try:
                        if step is not None:
                            step = int(step)
                    except ValueError:
                        msg = "Step is too big"
                        raise _errmsg(
                            msg, query, match.start(3), match.end(3),
                        ) from None

                    key = slice(start, stop, step)
                    nodes = [(target, key) for target in targets]
                elif match := _match_idx(query, end):
                    end = match.end()
                    try:
                        key = int(match.group())
                    except ValueError:
                        msg = "Index is too big"
                        raise _errmsg(msg, query, match.start(), end) from None

                    nodes = [(target, key) for target in targets]
                elif query[end:end + 1] == "'":
                    key, end = _scan_query_string(query, end + 1)
                    nodes = [(target, key) for target in targets]
                elif relative:
                    msg = "Filters are not allowed in relative query"
                    raise _errmsg(msg, query, end)
                else:
                    nodes = [
                        (target, key)
                        for target in targets
                        for key in (
                            target.keys()
                            if isinstance(target, dict) else
                            range(len(target))
                        )
                    ]
                    nodes, end = self._apply_filter(nodes, query, end)

                if query[end:end + 1] != "]":
                    msg = "Expecting a closing bracket"
                    raise _errmsg(msg, query, end)

                end += 1
            else:
                for target, key in nodes:
                    _check_query_key(target, key, allow_slice=allow_slice)

                return nodes, end

    def _paste_values(
        self,
        current_nodes: list[_Node],
        values: list[Any],
        operation: _Operation,
    ) -> None:
        if (mode := operation["mode"]) == "append":
            dst: str = operation.get("to", "@")
            dst_nodes: list[_Node] = self.select_nodes(
                current_nodes, dst, relative=True,
            )
            for (target, key), value in zip(dst_nodes, values):
                list.append(target[key], value)  # type: ignore
        elif mode == "extend":
            dst = operation.get("to", "@")
            dst_nodes = self.select_nodes(current_nodes, dst, relative=True)
            for (target, key), value in zip(dst_nodes, values):
                list.extend(target[key], value)  # type: ignore
        elif mode == "insert":
            dst = operation["to"]
            dst_nodes = self.select_nodes(current_nodes, dst, relative=True)

            # Future proofing: Reverse to preserve indices for filters
            for (current_target, _current_key), (target, key), value in zip(
                current_nodes[::-1], dst_nodes[::-1], values[::-1],
            ):
                if target is current_target:
                    msg: str = "Can not insert at current object"
                    raise ValueError(msg)

                list.insert(target, key, value)  # type: ignore
        elif mode == "set":
            dst = operation.get("to", "@")
            dst_nodes = self.select_nodes(
                current_nodes, dst, allow_slice=True, relative=True,
            )
            for (target, key), value in zip(dst_nodes, values):
                target[key] = value
        elif mode == "update":
            dst = operation.get("to", "@")
            dst_nodes = self.select_nodes(current_nodes, dst, relative=True)
            for (target, key), value in zip(dst_nodes, values):
                dict.update(target[key], value)  # type: ignore
        else:
            msg = "Unknown mode"
            raise ValueError(msg)

    def _apply_patch(
        self, root: list[Any], operations: list[_Operation],
    ) -> None:
        node: _Node = root, 0
        for operation in operations:
            if (op := operation["op"]) == "append":
                path: str = operation.get("path", "$")
                value: Any = operation["value"]
                for target, key in self.select_nodes(node, path):
                    list.append(target[key], deepcopy(value))  # type: ignore
            elif op == "assert":
                path = operation.get("path", "$")
                expr: str = operation["expr"]
                msg: str = operation.get("msg", f"Path {path}: {expr}")
                current_nodes: list[_Node] = self.select_nodes(node, path)
                if current_nodes != self.apply_filter(current_nodes, expr):
                    raise AssertionError(msg)
            elif op == "clear":
                path = operation.get("path", "$")
                for target, key in self.select_nodes(node, path):
                    new_target: Any = target[key]
                    if not isinstance(new_target, (dict, list)):
                        msg = (
                            "Target must be dict or list, not "
                            f"{type(new_target).__name__}"
                        )
                        raise TypeError(msg)

                    new_target.clear()
            elif op == "copy":
                path = operation.get("path", "$")
                src: str = operation["from"]
                current_nodes = self.select_nodes(node, path)
                values: list[Any] = [
                    deepcopy(target[key])
                    for target, key in self.select_nodes(
                        current_nodes, src, allow_slice=True, relative=True,
                    )
                ]
                self.paste_values(current_nodes, values, operation)
            elif op == "del":
                path = operation["path"]

                # Reverse to preserve indices for filters
                for target, key in self.select_nodes(
                    node, path, allow_slice=True,
                )[::-1]:
                    if target is root:
                        msg = "Can not delete root"
                        raise ValueError(msg)

                    del target[key]
            elif op == "extend":
                path = operation.get("path", "$")
                values = operation["values"]
                for target, key in self.select_nodes(node, path):
                    list.extend(target[key], deepcopy(values))  # type: ignore
            elif op == "insert":
                path = operation["path"]
                value = operation["value"]

                # Reverse to preserve indices for filters
                for target, key in self.select_nodes(node, path)[::-1]:
                    if target is root:
                        msg = "Can not insert at root"
                        raise ValueError(msg)

                    list.insert(target, key, deepcopy(value))  # type: ignore
            elif op == "move":
                path = operation.get("path", "$")
                src = operation["from"]
                current_nodes = self.select_nodes(node, path)
                src_nodes: list[_Node] = self.select_nodes(
                    current_nodes, src, allow_slice=True, relative=True,
                )
                values = []

                # Future proofing: Reverse to preserve indices for filters
                for (current_target, _current_key), (target, key) in zip(
                    current_nodes[::-1], src_nodes[::-1],
                ):
                    if target is current_target:
                        msg = "Can not move current object"
                        raise ValueError(msg)

                    values.append(target[key])
                    del target[key]

                # Undo reverse
                self.paste_values(current_nodes, values[::-1], operation)
            elif op == "reverse":
                path = operation.get("path", "$")
                for target, key in self.select_nodes(node, path):
                    list.reverse(target[key])  # type: ignore
            elif op == "set":
                path = operation.get("path", "$")
                value = operation["value"]
                for target, key in self.select_nodes(
                    node, path, allow_slice=True,
                ):
                    target[key] = deepcopy(value)
            elif op == "sort":
                path = operation.get("path", "$")
                reverse: bool = operation.get("reverse", False)
                for target, key in self.select_nodes(node, path):
                    list.sort(target[key], reverse=reverse)  # type: ignore
            elif op == "update":
                path = operation.get("path", "$")
                properties: dict[Any, Any] = operation["properties"]
                for target, key in self.select_nodes(node, path):
                    dict.update(  # type: ignore
                        target[key], deepcopy(properties),
                    )
            else:
                msg = "Unknown operation"
                raise ValueError(msg)

    def apply_patch(
        self, obj: Any, patch: _Operation | list[_Operation],
    ) -> Any:
        """Apply a JSON patch to a Python object.

        :param obj: a Python object
        :param patch: a :doc:`JSON patch </json-patch-spec>`
        :raises AssertionError: if an assertion fails
        :raises IndexError: if an index is out of range
        :raises JSONSyntaxError: if a query is invalid
        :raises KeyError: if a key is not found
        :raises TypeError: if a value has the wrong type
        :raises ValueError: if a value is invalid
        :return: the patched Python object

        Example:
            >>> import jsonyx as json
            >>> manipulator = json.Manipulator()
            >>> manipulator.apply_patch([1, 2, 3], {"op": "del", "path": "$[1]"})
            [1, 3]

        .. tip:: Using a :ref:`filter` instead of an index is more robust.

        """
        root: list[Any] = [obj]
        if isinstance(patch, dict):
            patch = [patch]

        self._apply_patch(root, patch)
        return root[0]

    def paste_values(
        self,
        current_nodes: _Node | list[_Node],
        values: Any | list[Any],
        operation: _Operation,
    ) -> None:
        """Paste value to a node or values to a list of nodes.

        :param current_nodes: a node or a list of nodes
        :param values: a value or a list of values
        :param operation: a JSON :ref:`copy` or :ref:`move` operation
        :raises IndexError: if an index is out of range
        :raises JSONSyntaxError: if a query is invalid
        :raises KeyError: if a key is not found
        :raises TypeError: if a value has the wrong type
        :raises ValueError: if a value is invalid

        Example:
            >>> import jsonyx as json
            >>> manipulator = json.Manipulator()
            >>> obj = [1, 2, 3]
            >>> root = [obj]
            >>> node = root, 0  # pointer to obj
            >>> manipulator.paste_values(node, 4, {"mode": "append"})
            >>> root[0]
            [1, 2, 3, 4]

        """
        if isinstance(current_nodes, tuple):
            current_nodes = [current_nodes]
            values = [values]

        self._paste_values(current_nodes, values, operation)

    def select_nodes(
        self,
        nodes: _Node | list[_Node],
        query: str,
        *,
        allow_slice: bool = False,
        relative: bool = False,
    ) -> list[_Node]:
        """Select nodes from a node or a list of nodes.

        :param nodes: a node or a list of nodes
        :param query: a :ref:`JSON select query <query>`
        :param allow_slice: allow slice
        :param relative: query must start with ``"@"`` instead of ``"$"``
        :raises IndexError: if an index is out of range
        :raises JSONSyntaxError: if the select query is invalid
        :raises KeyError: if a key is not found
        :raises TypeError: if a value has the wrong type
        :return: the selected list of nodes

        Example:
            >>> import jsonyx as json
            >>> manipulator = json.Manipulator()
            >>> obj = [1, 2, 3, 4, 5, 6]
            >>> root = [obj]
            >>> node = root, 0  # pointer to obj
            >>> for target, key in manipulator.select_nodes(node, "$[@ > 3]"):
            ...     target[key] = None
            ...
            >>> root[0]
            [1, 2, 3, None, None, None]

        """
        if isinstance(nodes, tuple):
            nodes = [nodes]

        nodes, end = self._select_nodes(
            nodes,
            query,
            allow_slice=allow_slice,
            relative=relative,
        )

        if end < len(query):
            msg: str = "Expecting end of file"
            raise _errmsg(msg, query, end)

        return nodes

    def apply_filter(
        self, nodes: _Node | list[_Node], query: str,
    ) -> list[_Node]:
        """Apply a JSON filter query to a node or a list of nodes.

        :param nodes: a node or a list of nodes
        :param query: a :ref:`JSON filter query <filter>`
        :raises IndexError: if an index is out of range
        :raises JSONSyntaxError: if the filter query is invalid
        :raises KeyError: if a key is not found
        :raises TypeError: if a value has the wrong type
        :return: the filtered list of nodes

        Example:
            >>> import jsonyx as json
            >>> manipulator = json.Manipulator()
            >>> obj = None
            >>> root = [obj]
            >>> node = root, 0  # pointer to obj
            >>> assert manipulator.apply_filter(node, "@ == null")

        """
        if isinstance(nodes, tuple):
            nodes = [nodes]

        nodes, end = self._apply_filter(nodes, query, 0)
        if end < len(query):
            msg: str = "Expecting end of file"
            raise _errmsg(msg, query, end)

        return nodes

    def load_query_value(self, s: str) -> Any:
        """Deserialize a JSON query value to a Python object.

        :param s: a :ref:`JSON query value <query_value>`
        :raises JSONSyntaxError: if the query value is invalid
        :return: a Python object

        Example:
            >>> import jsonyx as json
            >>> manipulator = json.Manipulator()
            >>> manipulator.load_query_value("'~'foo'")
            "'foo"

        """
        obj, end = self._scan_query_value(s)
        if end < len(s):
            msg: str = "Expecting end of file"
            raise _errmsg(msg, s, end)

        return obj


Manipulator.__module__ = "jsonyx"



"""JSON paste_values tests."""
from __future__ import annotations

__all__: list[str] = []

from typing import TYPE_CHECKING, Any

import pytest

from jsonyx import paste_values

if TYPE_CHECKING:
    _Operation = dict[str, Any]


def _apply_paste(obj: Any, value: Any, operation: _Operation) -> Any:
    root: list[Any] = [obj]
    paste_values((root, 0), value, operation)
    return root[0]


@pytest.mark.parametrize(("obj", "kwargs", "expected"), [
    ([1, 2, 3], {}, [1, 2, 3, 4]),
    ([[1, 2, 3]], {"to": "@[0]"}, [[1, 2, 3, 4]]),
])
def test_append(obj: Any, kwargs: _Operation, expected: Any) -> None:
    """Test append."""
    assert _apply_paste(obj, 4, {"mode": "append", **kwargs}) == expected


@pytest.mark.parametrize(("obj", "kwargs", "expected"), [
    ([1, 2, 3], {}, [1, 2, 3, 4, 5, 6]),
    ([[1, 2, 3]], {"to": "@[0]"}, [[1, 2, 3, 4, 5, 6]]),
])
def test_extend(obj: Any, kwargs: _Operation, expected: Any) -> None:
    """Test extend."""
    value: list[Any] = [4, 5, 6]
    assert _apply_paste(obj, value, {"mode": "extend", **kwargs}) == expected


def test_insert() -> None:
    """Test insert."""
    operation: _Operation = {"mode": "insert", "to": "@[0]"}
    assert _apply_paste([1, 2, 3], 0, operation) == [0, 1, 2, 3]


def test_insert_current_object() -> None:
    """Test insert at current object."""
    with pytest.raises(ValueError, match="Can not insert at current object"):
        _apply_paste(0, 0, {"mode": "insert", "to": "@"})


@pytest.mark.parametrize(("obj", "value", "kwargs", "expected"), [
    # Normal
    (0, 1, {}, 1),
    ([0], 1, {"to": "@[0]"}, [1]),
    ({"a": 0}, 1, {"to": "@.a"}, {"a": 1}),

    # Allow slice
    ([1, 2, 3], [4, 5, 6], {"to": "@[:]"}, [4, 5, 6]),
])
def test_set(obj: Any, value: Any, kwargs: _Operation, expected: Any) -> None:
    """Test set."""
    assert _apply_paste(obj, value, {"mode": "set", **kwargs}) == expected


@pytest.mark.parametrize(("obj", "kwargs", "expected"), [
    ({"a": 1, "b": 2, "c": 3}, {}, {"a": 4, "b": 5, "c": 6}),
    ([{"a": 1, "b": 2, "c": 3}], {"to": "@[0]"}, [{"a": 4, "b": 5, "c": 6}]),
])
def test_update(obj: Any, kwargs: _Operation, expected: Any) -> None:
    """Test update."""
    value: dict[str, Any] = {"a": 4, "b": 5, "c": 6}
    assert _apply_paste(obj, value, {"mode": "update", **kwargs}) == expected


@pytest.mark.parametrize("mode", [
    "assert", "clear", "copy", "del", "move", "reverse", "sort",
])
def test_unknown_mode(mode: str) -> None:
    """Test unknown mode."""
    with pytest.raises(ValueError, match="Unknown mode"):
        _apply_paste(0, 0, {"mode": mode})



"""jsonyx setup."""
from __future__ import annotations

__all__: list[str] = []

from pathlib import Path

# pylint: disable-next=E0401
from setuptools import Extension, setup  # type: ignore

if __name__ == "__main__":
    setup(
        name="jsonyx",
        version="2.3.0",
        description="Customizable JSON library for Python",
        long_description=Path("README.md").read_text(encoding="utf-8"),
        long_description_content_type="text/markdown",
        author="Nice Zombies",
        author_email="nineteendo19d0@gmail.com",
        maintainer="Nice Zombies",
        maintainer_email="nineteendo19d0@gmail.com",
        packages=["jsonyx", "jsonyx.test"],
        ext_modules=[
            Extension("_jsonyx", ["src/jsonyx/_speedups.c"], optional=True),
        ],
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
            "Programming Language :: Python :: 3.13",
            "Programming Language :: Python :: 3.14",
        ],
        license="GPL-3.0",
        keywords=["python", "json", "json-parser"],
        package_dir={"": "src"},
        # setuptools arguments
        entry_points={"console_scripts": ["jsonyx = jsonyx.__main__:main"]},
        python_requires=">=3.8",
        project_urls={
            "Homepage": "https://github.com/nineteendo/jsonyx",
            "Changelog": (
                "https://jsonyx.readthedocs.io/en/latest/changelog.html"
            ),
            "Documentation": "https://jsonyx.readthedocs.io",
            "Issues": "https://github.com/nineteendo/jsonyx/issues",
            "Sponser": "https://paypal.me/nineteendo",
        },
    )



#!/usr/bin/env python
"""A command line utility to manipulate JSON files."""
from __future__ import annotations

__all__: list[str] = ["main"]

import sys
from argparse import ArgumentParser
from decimal import Decimal
from sys import modules, stderr, stdin
from traceback import format_exception_only
from typing import Any, Literal, cast

from jsonyx import (
    Decoder, Encoder, Manipulator, TruncatedSyntaxError, __version__,
    format_syntax_error, make_patch,
)
from jsonyx.allow import EVERYTHING, NOTHING


# pylint: disable-next=R0903
class _Namespace:
    command: Literal["format", "patch", "diff"] | None
    commas: bool
    compact: bool
    ensure_ascii: bool
    indent: int | str | None
    indent_leaves: bool
    input_filename: str | None
    max_indent_level: int | None
    quoted_keys: bool
    nonstrict: bool
    output_filename: str | None
    sort_keys: bool
    trailing_comma: bool
    use_decimal: bool


# pylint: disable-next=R0903
class _DiffNameSpace(_Namespace):
    old_input_filename: str


# pylint: disable-next=R0903
class _PatchNameSpace(_Namespace):
    patch_filename: str


def _configure(parser: ArgumentParser) -> None:
    parser.add_argument("-v", "--version", action="version", version=(
        f"jsonyx {__version__} "
        f"({'C extension' if '_jsonyx' in modules else 'Python'})"
    ))
    parent_parser: ArgumentParser = ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "-a",
        "--ensure-ascii",
        action="store_true",
        help="escape non-ASCII characters",
    )
    parent_parser.add_argument(
        "-c",
        "--compact",
        action="store_true",
        help='avoid unnecessary whitespace after "," and ":"',
    )
    comma_group = parent_parser.add_mutually_exclusive_group()
    comma_group.add_argument(
        "-C",
        "--no-commas",
        action="store_false",
        dest="commas",
        help="don't separate items by commas when indented",
    )
    parent_parser.add_argument(
        "-d",
        "--use-decimal",
        action="store_true",
        help="use decimal instead of float",
    )
    indent_group = parent_parser.add_mutually_exclusive_group()
    indent_group.add_argument(
        "-i",
        "--indent",
        type=int,
        help="indent using the specified number of spaces",
        metavar="SPACES",
    )
    parent_parser.add_argument(
        "-l",
        "--no-indent-leaves",
        action="store_false",
        dest="indent_leaves",
        help="don't indent leaf objects and arrays",
    )
    parent_parser.add_argument(
        "-L",
        "--max-indent-level",
        type=int,
        help="the level up to which to indent",
        metavar="LEVEL",
    )
    parent_parser.add_argument(
        "-s",
        "--sort-keys",
        action="store_true",
        help="sort the keys of objects",
    )
    parent_parser.add_argument(
        "-S",
        "--nonstrict",
        action="store_true",
        help="allow all JSON deviations provided by jsonyx",
    )
    comma_group.add_argument(
        "-t",
        "--trailing-comma",
        action="store_true",
        help="add a trailing comma when indented",
    )
    indent_group.add_argument(
        "-T",
        "--indent-tab",
        action="store_const",
        const="\t",
        dest="indent",
        help="indent using tabs",
    )
    parent_parser.add_argument(
        "-q",
        "--unquoted-keys",
        action="store_false",
        dest="quoted_keys",
        help="don't quote keys which are identifiers",
    )
    commands = parser.add_subparsers(title="commands", dest="command")

    format_parser = commands.add_parser(
        "format",
        help="re-format or validate a JSON file",
        description="re-format or validate a JSON file",
        parents=[parent_parser],
    )
    format_parser.add_argument(
        "input_filename",
        nargs="?",
        help='the path to the input JSON file, or "-" for standard input',
    )
    format_parser.add_argument(
        "output_filename",
        nargs="?",
        help="the path to the output JSON file",
    )

    patch_parser = commands.add_parser(
        "patch",
        help="apply a JSON patch to the input file",
        description="apply a JSON patch to the input file",
        parents=[parent_parser],
    )
    patch_parser.add_argument(
        "patch_filename", help="the path to the JSON patch file",
    )
    patch_parser.add_argument(
        "input_filename",
        nargs="?",
        help='the path to the input JSON file, or "-" for standard input',
    )
    patch_parser.add_argument(
        "output_filename",
        nargs="?",
        help="the path to the output JSON file",
    )

    diff_parser = commands.add_parser(
        "diff",
        help="compare two JSON files and generate a diff in JSON patch format",
        description="compare two JSON files and generate a diff in JSON patch "
                    "format",
        parents=[parent_parser],
    )
    diff_parser.add_argument(
        "old_input_filename", help="the path to the old input JSON file",
    )
    diff_parser.add_argument(
        "input_filename",
        nargs="?",
        help='the path to the input JSON file, or "-" for standard input',
    )
    diff_parser.add_argument(
        "output_filename",
        nargs="?",
        help="the path to the output JSON file",
    )


def _run(args: _Namespace) -> None:
    decoder: Decoder = Decoder(
        allow=EVERYTHING if args.nonstrict else NOTHING,
        hooks={"float": Decimal if args.use_decimal else float},
    )
    encoder: Encoder = Encoder(
        allow=EVERYTHING if args.nonstrict else NOTHING,
        commas=args.commas,
        ensure_ascii=args.ensure_ascii,
        indent=args.indent,
        indent_leaves=args.indent_leaves,
        max_indent_level=args.max_indent_level,
        quoted_keys=args.quoted_keys,
        separators=(",", ":") if args.compact else (", ", ": "),
        sort_keys=args.sort_keys,
        trailing_comma=args.trailing_comma,
        types={"float": Decimal},
    )
    manipulator: Manipulator = Manipulator(
        allow=EVERYTHING if args.nonstrict else NOTHING,
        use_decimal=args.use_decimal,
    )
    try:
        if args.input_filename and args.input_filename != "-":
            input_obj: object = decoder.read(args.input_filename)
        elif stdin.isatty():
            input_obj = decoder.loads(
                "\n".join(iter(input, "")), filename="<stdin>",
            )
        else:
            input_obj = decoder.load(stdin)

        if args.command == "format":
            output_obj: Any = input_obj
        elif args.command == "patch":
            args = cast("_PatchNameSpace", args)
            patch: Any = decoder.read(args.patch_filename)
            output_obj = manipulator.apply_patch(input_obj, patch)
        else:
            args = cast("_DiffNameSpace", args)
            old_input_obj: object = decoder.read(args.old_input_filename)
            output_obj = make_patch(old_input_obj, input_obj)
            if len(output_obj) == 1:
                output_obj = output_obj[0]
    except TruncatedSyntaxError as exc:
        stderr.write("".join(format_syntax_error(exc)))
        sys.exit(1)
    except (
        AssertionError, LookupError, OSError, RecursionError, TypeError,
        ValueError,
    ) as exc:
        stderr.write("".join(format_exception_only(None, exc)))
        sys.exit(1)

    if args.output_filename and args.output_filename != "-":
        encoder.write(output_obj, args.output_filename)
    else:
        encoder.dump(output_obj)


def main() -> None:
    """Start jsonyx."""
    parser: ArgumentParser = ArgumentParser(
        description="a command line utility to manipulate JSON files",
    )
    _configure(parser)
    args: _Namespace = parser.parse_args(namespace=_Namespace())
    if not args.command:
        parser.print_help()
    else:
        try:
            _run(args)
        except BrokenPipeError as exc:
            sys.exit(exc.errno)


if __name__ == "__main__":
    main()



"""JSON make_patch tests."""
from __future__ import annotations

__all__: list[str] = []

from decimal import Decimal
from math import nan
from typing import TYPE_CHECKING, Any

import pytest

from jsonyx import make_patch

if TYPE_CHECKING:
    _Operation = dict[str, Any]


@pytest.mark.parametrize("key", [
    # Empty string
    "",

    # First character
    "\x00", " ", "!", "$", "0", "\xb2", "\u0300", "\u037a", "\u0488",

    # Remaining characters
    "A\x00", "A ", "A!", "A$", "A\xb2", "A\u037a", "A\u0488",
])
def test_string(key: str) -> None:
    """Test string."""
    assert make_patch({key: 0}, {}) == [{"op": "del", "path": f"$['{key}']"}]


@pytest.mark.parametrize(("key", "expected"), [
    # Tilde escapes
    ("'", "~'"),
    ("~", "~~"),

    # Multiple characters
    ("foo~bar", "foo~~bar"),
])
def test_string_escapes(key: str, expected: str) -> None:
    """Test string escapes."""
    assert make_patch({key: 0}, {}) == [
        {"op": "del", "path": f"$['{expected}']"},
    ]


@pytest.mark.parametrize("key", [
    # First character
    "A", "_", "\u16ee", "\u1885", "\u2118",

    # Remaining characters
    "A0", "AA", "A_", "A\u0300", "A\u2118",
])
def test_property(key: str) -> None:
    """Test property."""
    assert make_patch({key: 0}, {}) == [{"op": "del", "path": f"$.{key}"}]


@pytest.mark.parametrize("obj", [
    # Dict
    {}, {"": 0},

    # List
    [], [0],

    # NaN
    nan, Decimal("nan"),

    # Basic
    "", 0, Decimal(0), 0.0, Decimal("0.0"), True, False, None,
])
def test_equal(obj: object) -> None:
    """Test equal."""
    assert not make_patch(obj, obj)


@pytest.mark.parametrize(("old", "new", "expected"), [
    # Different type
    ("", 0, [{"op": "set", "path": "$", "value": 0}]),

    # Dict
    ({"a": 1, "c": 3, "d": 4}, {"b": 2, "c": 5, "d": 4}, [
        {"op": "del", "path": "$.a"},
        {"op": "set", "path": "$.c", "value": 5},
        {"op": "set", "path": "$.b", "value": 2},
    ]),

    # List
    ([1, 2, 3], [2, 4, 5], [
        {"op": "del", "path": "$[0]"},
        {"op": "set", "path": "$[1]", "value": 4},
        {"op": "insert", "path": "$[2]", "value": 5},
    ]),

    # Replacing NaN
    (nan, 0.0, [{"op": "set", "path": "$", "value": 0.0}]),

    # Basic
    (1, 2, [{"op": "set", "path": "$", "value": 2}]),
])  # type: ignore
def test_not_equal(
    old: object, new: object, expected: list[_Operation],
) -> None:
    """Test not equal."""
    assert make_patch(old, new) == expected



"""JSON dumps tests."""
from __future__ import annotations

__all__: list[str] = []

from collections import UserDict, UserList, UserString
from datetime import datetime, timezone
from decimal import Decimal
from enum import Enum
from typing import TYPE_CHECKING, Any

import pytest

from jsonyx.allow import NAN_AND_INFINITY, NON_STR_KEYS, SURROGATES

if TYPE_CHECKING:
    from types import ModuleType

_APOLLO11: datetime = datetime(1969, 7, 20, 20, 17, 40, tzinfo=timezone.utc)
_CIRCULAR_DICT: dict[str, object] = {}
_CIRCULAR_DICT[""] = _CIRCULAR_DICT
_CIRCULAR_LIST: list[object] = []
_CIRCULAR_LIST.append(_CIRCULAR_LIST)


# pylint: disable-next=R0903
class _MyBool:
    def __bool__(self) -> bool:
        return False


class _FloatEnum(float, Enum):
    ZERO = 0.0


class _IntEnum(int, Enum):
    ZERO = 0


@pytest.mark.parametrize(("obj", "expected"), [
    (True, "true"),
    (False, "false"),
    (None, "null"),
])
def test_singletons(json: ModuleType, obj: bool | None, expected: str) -> None:
    """Test singletons."""
    assert json.dumps(obj, end="") == expected


@pytest.mark.parametrize("num", [0, 1])
@pytest.mark.parametrize("int_type", [Decimal, int])
def test_int(json: ModuleType, num: int, int_type: type) -> None:
    """Test integer."""
    types: dict[str, type] = {"int": int_type}
    assert json.dumps(int_type(num), end="", types=types) == str(num)


@pytest.mark.parametrize("float_type", [Decimal, float])
def test_rational_number(json: ModuleType, float_type: type) -> None:
    """Test rational number."""
    types: dict[str, type] = {"float": float_type}
    assert json.dumps(float_type("0.0"), end="", types=types) == "0.0"


@pytest.mark.parametrize("num", ["NaN", "Infinity", "-Infinity"])
@pytest.mark.parametrize("float_type", [Decimal, float])
def test_nan_and_infinity(
    json: ModuleType, num: str, float_type: type,
) -> None:
    """Test NaN and (negative) infinity."""
    allow: frozenset[str] = NAN_AND_INFINITY
    types: dict[str, type] = {"float": float_type}
    assert json.dumps(float_type(num), allow=allow, end="", types=types) == num


@pytest.mark.parametrize("num", ["NaN", "Infinity", "-Infinity"])
@pytest.mark.parametrize("float_type", [Decimal, float])
def test_nan_and_infinity_not_allowed(
    json: ModuleType, num: str, float_type: type,
) -> None:
    """Test NaN and (negative) infinity when not allowed."""
    with pytest.raises(ValueError, match="is not allowed"):
        json.dumps(float_type(num), types={"float": float_type})


@pytest.mark.parametrize("num", [
    # Quiet NaN
    "NaN2", "-NaN", "-NaN2",

    # Signaling NaN
    "sNaN", "sNaN2", "-sNaN", "-sNaN2",
])
def test_nan_payload(json: ModuleType, num: str) -> None:
    """Test NaN payload."""
    types: dict[str, type] = {"float": Decimal}
    with pytest.raises(ValueError, match="is not JSON serializable"):
        json.dumps(Decimal(num), allow=NAN_AND_INFINITY, types=types)


@pytest.mark.parametrize("obj", [
    # Key
    {_IntEnum.ZERO: 0}, {_FloatEnum.ZERO: 0},

    # Value
    _IntEnum.ZERO, _FloatEnum.ZERO,
])
def test_enum(json: ModuleType, obj: float | dict[object, object]) -> None:
    """Test enum."""
    with pytest.raises(ValueError, match="is not JSON serializable"):
        json.dumps(obj, allow=NON_STR_KEYS)


@pytest.mark.parametrize("obj", [
    # UTF-8
    "\xa3", "\u0418", "\u0939", "\u20ac", "\ud55c", "\U00010348", "\U001096b3",

    # Surrogates
    "\ud800", "\udf48",  # noqa: PT014
])
def test_string(json: ModuleType, obj: str) -> None:
    """Test string."""
    assert json.dumps(obj, end="") == f'"{obj}"'


@pytest.mark.parametrize(("obj", "expected"), [
    ("\xa3", r"\u00a3"),
    ("\u0418", r"\u0418"),
    ("\u0939", r"\u0939"),
    ("\u20ac", r"\u20ac"),
    ("\ud55c", r"\ud55c"),
    ("\U00010348", r"\ud800\udf48"),
    ("\U001096b3", r"\udbe5\udeb3"),
])
def test_string_ensure_ascii(
    json: ModuleType, obj: str, expected: str,
) -> None:
    """Test string with ensure_ascii."""
    assert json.dumps(obj, end="", ensure_ascii=True) == f'"{expected}"'


@pytest.mark.parametrize("obj", [
    # Empty string
    "",

    # One character
    "$",

    # Multiple characters
    "foo",
])
@pytest.mark.parametrize("ensure_ascii", [True, False])
def test_ascii_string(json: ModuleType, obj: str, ensure_ascii: bool) -> None:
    """Test ascii string."""
    assert json.dumps(obj, end="", ensure_ascii=ensure_ascii) == f'"{obj}"'


@pytest.mark.parametrize(("obj", "expected"), [
    # Control characters
    ("\x00", r"\u0000"),
    ("\x08", r"\b"),
    ("\t", r"\t"),
    ("\n", r"\n"),
    ("\x0c", r"\f"),
    ("\r", r"\r"),
    ('"', r"\""),
    ("\\", r"\\"),

    # Multiple characters
    (r"foo\bar", r"foo\\bar"),
])
@pytest.mark.parametrize("ensure_ascii", [True, False])
def test_control_characters(
    json: ModuleType, obj: str, ensure_ascii: bool, expected: str,
) -> None:
    """Test control characters."""
    s: str = json.dumps(obj, end="", ensure_ascii=ensure_ascii)
    assert s == f'"{expected}"'


@pytest.mark.parametrize(("obj", "expected"), [
    ("\ud800", r"\ud800"),
    ("\udf48", r"\udf48"),
])
def test_surrogate_escapes(json: ModuleType, obj: str, expected: str) -> None:
    """Test surrogate escapes."""
    s: str = json.dumps(obj, allow=SURROGATES, end="", ensure_ascii=True)
    assert s == f'"{expected}"'


@pytest.mark.parametrize("obj", ["\ud800", "\udf48"])  # noqa: PT014
def test_surrogate_escapes_not_allowed(json: ModuleType, obj: str) -> None:
    """Test surrogate escapes when not allowed."""
    with pytest.raises(ValueError, match="Surrogates are not allowed"):
        json.dumps(obj, ensure_ascii=True)


@pytest.mark.parametrize(("obj", "expected"), [
    # Empty list
    ([], "[]"),

    # One value
    ([0], "[0]"),

    # Multiple values
    ([1, 2, 3], "[1, 2, 3]"),
])  # type: ignore
def test_list(json: ModuleType, obj: list[object], expected: str) -> None:
    """Test list."""
    assert json.dumps(obj, end="") == expected


@pytest.mark.parametrize(("indent", "expected"), [
    # Integer
    (0, ""),
    (1, " "),

    # String
    ("\t", "\t"),
])
def test_list_indent(
    json: ModuleType, indent: int | str, expected: str,
) -> None:
    """Test list indent."""
    s: str = json.dumps([1, 2, 3], end="", indent=indent)
    assert s == f"[\n{expected}1,\n{expected}2,\n{expected}3\n]"


def test_empty_list_indent(json: ModuleType) -> None:
    """Test empty list indent."""
    assert json.dumps([], end="", indent=1) == "[]"


@pytest.mark.parametrize(("obj", "expected"), [
    ([1, 2, 3], "[1, 2, 3]"),
    ([[1, 2, 3]], "[\n [1, 2, 3]\n]"),
])
def test_list_no_indent_leaves(
    json: ModuleType, obj: list[object], expected: str,
) -> None:
    """Test list indent without indent_leaves."""
    assert json.dumps(obj, end="", indent=1, indent_leaves=False) == expected


def test_list_max_indent_level(json: ModuleType) -> None:
    """Test list indent with max_indent_level."""
    s: str = json.dumps([[1, 2, 3]], end="", indent=1, max_indent_level=1)
    assert s == "[\n [1, 2, 3]\n]"


def test_list_recursion(json: ModuleType) -> None:
    """Test list recursion."""
    obj: list[object] = []
    for _ in range(100_000):
        obj = [obj]

    with pytest.raises(RecursionError):
        json.dumps(obj)


def test_array_types(json: ModuleType) -> None:
    """Test array_types."""
    obj: UserList[object] = UserList([1, 2, 3])
    assert json.dumps(obj, end="", types={"array": UserList}) == "[1, 2, 3]"


@pytest.mark.parametrize(("obj", "expected"), [
    # Empty dict
    ({}, "{}"),

    # One value
    ({"": 0}, '{"": 0}'),

    # Multiple values
    ({"a": 1, "b": 2, "c": 3}, '{"a": 1, "b": 2, "c": 3}'),
])
def test_dict(
    json: ModuleType, obj: dict[str, object], expected: str,
) -> None:
    """Test dict."""
    assert json.dumps(obj, end="") == expected


@pytest.mark.parametrize("key", [
    # First character
    "\xb2", "\u0300", "\u037a", "\u0488",

    # Remaining characters
    "A\xb2", "A\u037a", "A\u0488",
])
def test_quoted_keys(json: ModuleType, key: str) -> None:
    """Test quoted keys."""
    assert json.dumps({key: 0}, end="", quoted_keys=False) == f'{{"{key}": 0}}'


@pytest.mark.parametrize(("key", "expected"), [
    # First character
    ("\xb2", r"\u00b2"),
    ("\u0300", r"\u0300"),
    ("\u037a", r"\u037a"),
    ("\u0488", r"\u0488"),
    ("\u16ee", r"\u16ee"),
    ("\u1885", r"\u1885"),
    ("\u2118", r"\u2118"),

    # Remaining characters
    ("A\xb2", r"A\u00b2"),
    ("A\u0300", r"A\u0300"),
    ("A\u037a", r"A\u037a"),
    ("A\u0488", r"A\u0488"),
    ("A\u2118", r"A\u2118"),
])
def test_quoted_keys_ensure_ascii(
    json: ModuleType, key: str, expected: str,
) -> None:
    """Test quoted keys with ensure_ascii."""
    s: str = json.dumps({key: 0}, end="", ensure_ascii=True, quoted_keys=False)
    assert s == f'{{"{expected}": 0}}'


@pytest.mark.parametrize("key", [
    # Empty string
    "",

    # First character
    " ", "!", "$", "0",

    # Remaining characters
    "A ", "A!", "A$",
])
@pytest.mark.parametrize("ensure_ascii", [True, False])
def test_quoted_ascii_keys(
    json: ModuleType, key: str, ensure_ascii: bool,
) -> None:
    """Test quoted ascii keys."""
    assert json.dumps(
        {key: 0}, end="", ensure_ascii=ensure_ascii, quoted_keys=False,
    ) == f'{{"{key}": 0}}'


@pytest.mark.parametrize(("key", "expected"), [
    # First character
    ("\x00", r"\u0000"),

    # Remaining characters
    ("A\x00", r"A\u0000"),
])
@pytest.mark.parametrize("ensure_ascii", [True, False])
def test_control_characters_key(
    json: ModuleType, key: str, ensure_ascii: bool, expected: str,
) -> None:
    """Test control characters in key."""
    assert json.dumps(
        {key: 0}, end="", ensure_ascii=ensure_ascii, quoted_keys=False,
    ) == f'{{"{expected}": 0}}'


@pytest.mark.parametrize("key", [
    # First character
    "\u16ee", "\u1885", "\u2118",

    # Remaining characters
    "A\u0300", "A\u2118",
])
def test_unquoted_keys(json: ModuleType, key: str) -> None:
    """Test unquoted keys."""
    assert json.dumps({key: 0}, end="", quoted_keys=False) == f"{{{key}: 0}}"


@pytest.mark.parametrize("key", [
    # First character
    "A", "_",

    # Remaining characters
    "A0", "AA", "A_",
])
@pytest.mark.parametrize("ensure_ascii", [True, False])
def test_unquoted_ascii_keys(
    json: ModuleType, key: str, ensure_ascii: bool,
) -> None:
    """Test unquoted ascii keys."""
    assert json.dumps(
        {key: 0}, end="", ensure_ascii=ensure_ascii, quoted_keys=False,
    ) == f"{{{key}: 0}}"


@pytest.mark.parametrize(("key", "expected"), [
    (True, "true"),
    (False, "false"),
    (None, "null"),
])
def test_singleton_keys(
    json: ModuleType, key: bool | None, expected: str,
) -> None:
    """Test singleton keys."""
    s: str = json.dumps({key: 0}, allow=NON_STR_KEYS, end="")
    assert s == f'{{"{expected}": 0}}'


@pytest.mark.parametrize("int_type", [Decimal, int])
def test_int_key(json: ModuleType, int_type: type) -> None:
    """Test int key."""
    obj: dict[object, object] = {int_type(0): 0}
    types: dict[str, type] = {"int": int_type}
    s: str = json.dumps(obj, allow=NON_STR_KEYS, end="", types=types)
    assert s == '{"0": 0}'


@pytest.mark.parametrize("float_type", [Decimal, float])
def test_float_key(json: ModuleType, float_type: type) -> None:
    """Test float key."""
    obj: dict[object, object] = {float_type("0.0"): 0}
    types: dict[str, type] = {"float": float_type}
    s: str = json.dumps(obj, allow=NON_STR_KEYS, end="", types=types)
    assert s == '{"0.0": 0}'


@pytest.mark.parametrize("key", [0, 0.0, True, False, None])
def test_non_str_keys_not_allowed(json: ModuleType, key: object) -> None:
    """Test non-string keys if not allowed."""
    with pytest.raises(TypeError, match="Non-string keys are not allowed"):
        json.dumps({key: 0})


@pytest.mark.parametrize("key", [
    # JSON values
    (),

    # No JSON values
    b"", 0j, frozenset(), memoryview(b""), object(),
])  # type: ignore
def test_unserializable_key(json: ModuleType, key: object) -> None:
    """Test unserializable key."""
    with pytest.raises(TypeError, match="Keys must be str, not"):
        json.dumps({key: 0})


@pytest.mark.parametrize("key", [
    # JSON values
    (), 0, 0.0, True, False, None,

    # No JSON values
    b"", 0j, frozenset(), memoryview(b""), object(),
])  # type: ignore
def test_skip_keys(json: ModuleType, key: object) -> None:
    """Test skipkeys."""
    assert json.dumps({key: 0}, end="", indent=1, skipkeys=True) == "{}"


def test_sort_keys(json: ModuleType) -> None:
    """Test sort_keys."""
    s: str = json.dumps({"c": 3, "b": 2, "a": 1}, end="", sort_keys=True)
    assert s == '{"a": 1, "b": 2, "c": 3}'


@pytest.mark.parametrize(("indent", "expected"), [
    # Integer
    (0, ""),
    (1, " "),

    # String
    ("\t", "\t"),
])
def test_dict_indent(
    json: ModuleType, indent: int | str, expected: str,
) -> None:
    """Test dict indent."""
    obj: dict[str, object] = {"a": 1, "b": 2, "c": 3}
    s: str = json.dumps(obj, end="", indent=indent)
    assert s == (
        f'{{\n{expected}"a": 1,\n{expected}"b": 2,\n{expected}"c": 3\n}}'
    )


def test_empty_dict_indent(json: ModuleType) -> None:
    """Test empty dict indent."""
    assert json.dumps({}, end="", indent=1) == "{}"


@pytest.mark.parametrize(("obj", "expected"), [
    ({"a": 1, "b": 2, "c": 3}, '{"a": 1, "b": 2, "c": 3}'),
    ({"": {"a": 1, "b": 2, "c": 3}}, '{\n "": {"a": 1, "b": 2, "c": 3}\n}'),
])
def test_dict_no_indent_leaves(
    json: ModuleType, obj: dict[str, object], expected: str,
) -> None:
    """Test dict indent without indent_leaves."""
    assert json.dumps(obj, end="", indent=1, indent_leaves=False) == expected


def test_dict_max_indent_level(json: ModuleType) -> None:
    """Test dict indent with max_indent_level."""
    obj: dict[str, object] = {"": {"a": 1, "b": 2, "c": 3}}
    s: str = json.dumps(obj, end="", indent=1, max_indent_level=1)
    assert s == '{\n "": {"a": 1, "b": 2, "c": 3}\n}'


def test_dict_recursion(json: ModuleType) -> None:
    """Test dict recursion."""
    obj: dict[str, object] = {}
    for _ in range(100_000):
        obj = {"": obj}

    with pytest.raises(RecursionError):
        json.dumps(obj)


def test_object_types(json: ModuleType) -> None:
    """Test object_types."""
    obj: UserDict[str, object] = UserDict({"a": 1, "b": 2, "c": 3})
    s: str = json.dumps(obj, end="", types={"object": UserDict})
    assert s == '{"a": 1, "b": 2, "c": 3}'


@pytest.mark.parametrize("obj", [
    b"", 0j, bytearray(), frozenset(), memoryview(b""), object(), range(0),
    set(), slice(0),
])  # type: ignore
def test_unserializable_value(json: ModuleType, obj: object) -> None:
    """Test unserializable value."""
    with pytest.raises(TypeError, match="is not JSON serializable"):
        json.dumps(obj)


@pytest.mark.parametrize(("obj", "expected"), [
    (_MyBool(), "false"),
    ({_MyBool(): 0}, '{"false": 0}'),
])
def test_bool_types(
    json: ModuleType, obj: _MyBool | dict[object, object], expected: str,
) -> None:
    """Test bool_types."""
    types: dict[str, type] = {"bool": _MyBool}
    assert json.dumps(obj, end="", allow=NON_STR_KEYS, types=types) == expected


@pytest.mark.parametrize(("obj", "expected"), [
    (UserString(""), '""'),
    ({UserString(""): 0}, '{"": 0}'),
])
def test_str_types(
    json: ModuleType, obj: UserString | dict[object, object], expected: str,
) -> None:
    """Test str_types."""
    assert json.dumps(obj, end="", types={"str": UserString}) == expected


@pytest.mark.parametrize(("obj", "expected"), [
    (_APOLLO11, '"1969-07-20T20:17:40+00:00"'),
    ({_APOLLO11: 0}, '{"1969-07-20T20:17:40+00:00": 0}'),
])
def test_hook(
    json: ModuleType, obj: datetime | dict[object, object], expected: str,
) -> None:
    """Test hook."""
    def datetime_hook(obj: Any) -> Any:
        return obj.isoformat() if isinstance(obj, datetime) else obj

    assert json.dumps(obj, end="", hook=datetime_hook) == expected


@pytest.mark.parametrize(("obj", "expected"), [
    ([1 + 2j], '[\n {"real": 1.0, "imag": 2.0}\n]'),
    ({"": 1 + 2j}, '{\n "": {"real": 1.0, "imag": 2.0}\n}'),
])
def test_hook_no_indent_leaves(
    json: ModuleType, obj: complex | dict[object, object], expected: str,
) -> None:
    """Test hook without indent_leaves."""
    def complex_hook(obj: Any) -> Any:
        if isinstance(obj, complex):
            return {"real": obj.real, "imag": obj.imag}

        return obj

    assert json.dumps(
        obj,
        end="",
        hook=complex_hook,
        indent=1,
        indent_leaves=False,
    ) == expected


@pytest.mark.parametrize("obj", [_CIRCULAR_DICT, _CIRCULAR_LIST])
def test_circular_reference(
    json: ModuleType, obj: dict[str, object] | list[object],
) -> None:
    """Test circular reference."""
    with pytest.raises(ValueError, match="Unexpected circular reference"):
        json.dumps(obj)


@pytest.mark.parametrize("obj", [_CIRCULAR_DICT, _CIRCULAR_LIST])
def test_circular_reference_no_check_circular(
    json: ModuleType, obj: dict[str, object] | list[object],
) -> None:
    """Test circular reference without check_circular."""
    with pytest.raises(RecursionError):
        json.dumps(obj, check_circular=False)


@pytest.mark.parametrize("check_circular", [True, False])
@pytest.mark.parametrize(("obj", "expected"), [
    ([[]] * 3, "[[], [], []]"),
    ([{}] * 3, "[{}, {}, {}]"),
])  # type: ignore
def test_shadow_copy(
    json: ModuleType, obj: list[object], expected: str, check_circular: bool,
) -> None:
    """Test shadow copy."""
    assert json.dumps(obj, check_circular=check_circular, end="") == expected


@pytest.mark.parametrize(("obj", "expected"), [
    ([1, 2, 3], "[1, 2, 3]"),
    ({"a": 1, "b": 2, "c": 3}, '{"a": 1, "b": 2, "c": 3}'),
])
def test_no_commas(
    json: ModuleType, obj: dict[str, object] | list[object], expected: str,
) -> None:
    """Test no commas."""
    assert json.dumps(obj, commas=False, end="") == expected


@pytest.mark.parametrize(("obj", "expected"), [
    ([1, 2, 3], "[\n 1\n 2\n 3\n]"),
    ({"a": 1, "b": 2, "c": 3}, '{\n "a": 1\n "b": 2\n "c": 3\n}'),
])
@pytest.mark.parametrize("trailing_comma", [True, False])
def test_no_commas_indent(
    json: ModuleType,
    obj: dict[str, object] | list[object],
    expected: str,
    trailing_comma: bool,
) -> None:
    """Test no commas with indent."""
    assert json.dumps(
        obj, commas=False, end="", indent=1, trailing_comma=trailing_comma,
    ) == expected


def test_default_end(json: ModuleType) -> None:
    """Test default end."""
    assert json.dumps(0) == "0\n"


def test_custom_end(json: ModuleType) -> None:
    """Test custom end."""
    assert json.dumps(0, end="") == "0"


@pytest.mark.parametrize(("obj", "expected"), [
    ([1, 2, 3], "[1,2,3]"),
    ({"a": 1, "b": 2, "c": 3}, '{"a":1,"b":2,"c":3}'),
])
def test_separators(
    json: ModuleType, obj: dict[str, object] | list[object], expected: str,
) -> None:
    """Test separators."""
    assert json.dumps(obj, end="", separators=(",", ":")) == expected


@pytest.mark.parametrize(("obj", "expected"), [
    ([1, 2, 3], "[1, 2, 3]"),
    ({"a": 1, "b": 2, "c": 3}, '{"a": 1, "b": 2, "c": 3}'),
])
def test_trailing_comma(
    json: ModuleType, obj: dict[str, object] | list[object], expected: str,
) -> None:
    """Test trailing_comma."""
    assert json.dumps(obj, end="", trailing_comma=True) == expected


@pytest.mark.parametrize(("obj", "expected"), [
    ([1, 2, 3], "[\n 1,\n 2,\n 3,\n]"),
    ({"a": 1, "b": 2, "c": 3}, '{\n "a": 1,\n "b": 2,\n "c": 3,\n}'),
])
def test_trailing_comma_indent(
    json: ModuleType, obj: dict[str, object] | list[object], expected: str,
) -> None:
    """Test trailing_comma with indent."""
    assert json.dumps(obj, end="", indent=1, trailing_comma=True) == expected
